// Sections of this code have been adapted from RegridX.C.m4 of LPARX
// by Scott Kohn
//
#include <iostream.h>
#include <iomanip.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "lparx.h"
ifelse(NDIM,2,@
#include "cadez_hierarchy2.h"
@,NDIM,3,@
#include "cadez_hierarchy3.h"
@)

//declare_Array1(int)
#include "Array1int.h"

#define MAX(x,y) ((x) > (y) ? (x) : (y))
#define MIN(x,y) ((x) < (y) ? (x) : (y))
#define ROUND(a,s) (((a)+((s)/2))/(s))

#define NDUPFIELDS 1

ifelse(NDIM,2,@
#define I_X 0
#define I_Y 1
@,NDIM,3,@
#define I_W 0
#define I_X 1
#define I_Y 2
@)

#define RegionListNULL ((RegionList *) 0)
class RegionList {
public:
	RegionList *next;
	RegionX region;
};

#define FillListNULL ((FillList *) 0)
class FillList {
public:
	FillList *next;
	PointX where;
	int count;
	FillList(PointX pt) {
		where = pt;
		next = FillListNULL;
		count = 0;
	}
};

void printX(XArray1(GridX(double)) &);
ifelse(NDIM,2,@
extern "FORTRAN" void ctof3x3(double *, double *, int *, int *, int *, int *);
@,NDIM,3,@
extern "FORTRAN" void ctof3x3x3(double *, double *, int *, int *, int *, int *);
@)
extern "FORTRAN" int d2error(double *, double *, const double &,
										int *, int *, int*);

static RegionList *RegridLevel(int &, cadez_discretization &, 
				const RegionX &, const int, const PointX &);

void Balance_Partitions(RegionList *&, RegionX *&, int *&, int &,
								const int , const PointX &);

static void EnsureNesting(RegionList *&, int &, cadez_discretization &);

static int count_flagged(const RegionX &, cadez_discretization &);

static void SliceAndDice(RegionList *&, int &, const int, const PointX &);

static void SortRegions(RegionList *&, const int);

int cadez_hierarchy::adapt_refine(const int New_Level, const double trunc)
{
//
//------------------------------------------------------------------------------
//	Check the location of the new level in the hierarchy.  If the new level is
//	within the current hierarchy, save the old discretization for to initialize
//	the new discretization where ever they overlap.  If the new level is
//	outside of the current hierarchy, make sure it is only one level out.
//
	cadez_discretization *Old_Level = 0;
	if(New_Level < nlevels) {
		Old_Level = level[New_Level];
	} else if(New_Level > nlevels) {
		cerr << "Cannot create level " << New_Level << ", level " << New_Level-1
				<< " does not exist!" << endl;
		exit(1);
	} else {
		++nlevels;
	}
//
//------------------------------------------------------------------------------
	cadez_discretization & f = *(level[New_Level-1]);
//
//------------------------------------------------------------------------------
//	If this level has been previously refined, first set Refine_Mask to 1.0
	if(Old_Level != 0)
		for_all_1(l,f.TmpC)
			f.Refine_Mask(l).fill(1.0);
		end_for_all_no_sync
//
//------------------------------------------------------------------------------
//	Set Refine_Mask to 0.0 for points needing refinement.
	for_all_1(k,f.TmpC)
		d2error(f.PsiD(k).data(),f.Refine_Mask(k).data(),trunc,
					(int *)(f.extents+k),(int *)(f.interiors+k),(int *)(f.fills+k));
	end_for_all_no_sync
//
//------------------------------------------------------------------------------
//	Calculate the 3 bounding boxes of the refinement region
	RegionX bbox1, bbox2, bbox3;
	for_1(i,f.Refine_Mask)
		bbox1 += f.interiors[i]*f.domain_R1;
		bbox2 += f.interiors[i]*f.domain_R2;
		bbox3 += f.interiors[i]*f.domain_R3;
	end_for
//
//	Calculate refinement patches.
	int Nr1 = 0, Nr2 = 0, Nr3 = 0;
	PointX minsize(8);
	RegionList *list1 = RegridLevel(Nr1,f,bbox1,70,minsize);
	RegionList *list2 = RegridLevel(Nr2,f,bbox2,70,minsize);
	RegionList *list3 = RegridLevel(Nr3,f,bbox3,70,minsize);
//
//	Make sure that the new refinement regions are nested properly
	EnsureNesting(list1,Nr1,f);
	EnsureNesting(list2,Nr2,f);
	EnsureNesting(list3,Nr3,f);
//
//------------------------------------------------------------------------------
//	Remove any subregions with zero flagged points and consolidate lists
	int regions = Nr1 + Nr2 + Nr3;
	RegionList *final = RegionListNULL;
	while(list3 != RegionListNULL) {
		RegionList *check = list3;
		list3 = list3->next;
		if(count_flagged(check->region,f) != 0) {
			check->next = final;
			final = check;
		} else {
			delete check;
			regions--;
		}
	}
	while(list2 != RegionListNULL) {
		RegionList *check = list2;
		list2 = list2->next;
		if(count_flagged(check->region,f) != 0) {
			check->next = final;
			final = check;
		} else {
			delete check;
			regions--;
		}
	}
	while(list1 != RegionListNULL) {
		RegionList *check = list1;
		list1 = list1->next;
		if(count_flagged(check->region,f) != 0) {
			check->next = final;
			final = check;
		} else {
			delete check;
			regions--;
		}
	}
//
//------------------------------------------------------------------------------
// Return if the new level contains no points.
	if(regions == 0) return(0);
//------------------------------------------------------------------------------
// Create new ADAPTED level.
	// Compute Cadez domain sizes for new, refined level.
   const int nep = 2 * f.domain_R1.extents(I_Y);
   const int nem = 2 * f.domain_R2.extents(I_Y);
   const int ne3 = 2 * f.domain_R3.extents(I_Y);
   const int nsp = 2 * f.domain_R1.extents(I_X);
   const int nsm = 2 * f.domain_R2.extents(I_X);
ifelse(NDIM,3,@
   const int nphi = 2 * f.domain_R3.extents(I_W);
@)
// Allocate and initialize a new level.
   if(New_Level+1 < size) {  // Enough space, just allocate and initialize.
ifelse(NDIM,2,@
      level[New_Level] = new cadez_discretization(domain,Adapted,
																		nep,nem,ne3,nsp,nsm);
@,NDIM,3,@
      level[New_Level] = new cadez_discretization(domain,Adapted,
																		nep,nem,ne3,nsp,nsm,nphi);
@)
   } else { // Not enough space. Expand level, then allocate and initialize.
      const int tmpsize = size + resize_chunk;
      cadez_discretization ** tmplev = new cadez_discretization*[tmpsize];
      memset(tmplev, 0, sizeof(cadez_discretization*)*tmpsize);
      for(int i = 0; i < size; ++i)
         tmplev[i] = level[i];
ifelse(NDIM,2,@
      tmplev[New_Level] = new cadez_discretization(domain,Adapted,
																		nep,nem,ne3,nsp,nsm);
@,NDIM,3,@
      tmplev[New_Level] = new cadez_discretization(domain,Adapted,
																		nep,nem,ne3,nsp,nsm,nphi);
@)
      delete level;
      level = tmplev;
      size = tmpsize;
   }
//------------------------------------------------------------------------------
//
//	Balance the list of refinements across the processors and
//	set the fine and coarse interiors, extents, and fills regions.
	cadez_discretization & al = *level[New_Level];
	RegionX *cint_new = 0;
	int *map_new = 0;
	minsize *= 2;
	Balance_Partitions(final,cint_new,map_new,regions,mpNodes(),minsize);
//	Set old Refine_Mask to final state
	for_all_1(l,f.Refine_Mask)
		for(int m=0; m < regions; m++) {
			RegionX ref = f.interiors[l]*cint_new[m];
			GridX(double) & dst = f.Refine_Mask(l);
			BeginFastIndexNDIM(dst, double);
ifelse(NDIM,2,@
			for_2(i,j,ref)
				FastIndex2(dst,i,j) = 0.0;
			end_for
@,NDIM,3,@
			for_3(i,j,k,ref)
				FastIndex3(dst,i,j,k) = 0.0;
			end_for
@)
			EndFastIndexNDIM(dst);
		}
	end_for_all
//
	RegionX *int_new = new RegionX[regions];
	RegionX *ext_new = new RegionX[regions];
	RegionX *fil_new = new RegionX[regions];
	RegionX *cext_new = new RegionX[regions];
	RegionX *cfil_new = new RegionX[regions];
	RegionX *cpro_new = new RegionX[regions];
	RegionX *intseq = int_new;
	RegionX *extseq = ext_new;
	RegionX *filseq = fil_new;
	RegionX *cintseq = cint_new;
	RegionX *cextseq = cext_new;
	RegionX *cfilseq = cfil_new;
	RegionX *cproseq = cpro_new;
ifelse(NDIM,3,@
//	set up Point3 for adding periodicity to phi component of fill regions
	Point3 periodic(0); periodic(I_W) = 1;
@)
//	Set the regions
	for(int k = 0; k < regions; k++) {
		*intseq = RegionX(cintseq->lower()*2 - 1,cintseq->upper()*2);
		*filseq = grow(*intseq,1);
		if(!((*intseq)*al.domain_R1).empty()) {
ifelse(NDIM,2,@
			*filseq = (*filseq)*al.domain_R1;
@,NDIM,3,@
			*filseq = (*filseq)*grow(al.domain_R1,periodic);
@)
			if(intseq->upper(I_Y) == 0) {
				int filext = 4 - intseq->extents(I_Y);
				filseq->upper(I_Y) += (filext > 1) ? filext : 1;
			}
		} else if(!((*intseq)*al.domain_R2).empty()) {
ifelse(NDIM,2,@
			*filseq = (*filseq)*al.domain_R2;
@,NDIM,3,@
			*filseq = (*filseq)*grow(al.domain_R2,periodic);
@)
			if(intseq->upper(I_Y) == 0) {
				int filext = 4 - intseq->extents(I_Y);
				filseq->upper(I_Y) += (filext > 1) ? filext : 1;
			}
		} else {
ifelse(NDIM,2,@
			*filseq = (*filseq)*al.domain_R3;
@,NDIM,3,@
			*filseq = (*filseq)*grow(al.domain_R3,periodic);
@)
			if(intseq->lower(I_Y) == 1) filseq->lower(I_Y) -= 1;
		}
//
		*(extseq++) = grow(*(filseq++),1);
		intseq++;
//
		*cfilseq = *cextseq = grow(*cintseq,1);
		if(!((*cintseq)*f.domain_R1).empty()) {
ifelse(NDIM,2,@
			*cfilseq = (*cfilseq)*f.domain_R1;
@,NDIM,3,@
			*cfilseq = (*cfilseq)*grow(f.domain_R1,periodic);
@)
			*cproseq = *cfilseq;
			if(cintseq->upper(I_Y) == 0) {
				int filext = 3 - cintseq->extents(I_Y);
				cfilseq->upper(I_Y) += (filext > 1) ? filext : 1;
				cproseq->upper(I_Y) += 1;
			}
		} else if(!((*cintseq)*f.domain_R2).empty()) {
ifelse(NDIM,2,@
			*cfilseq = (*cfilseq)*f.domain_R2;
@,NDIM,3,@
			*cfilseq = (*cfilseq)*grow(f.domain_R2,periodic);
@)
			*cproseq = *cfilseq;
			if(cintseq->upper(I_Y) == 0) {
				int filext = 3 - cintseq->extents(I_Y);
				cfilseq->upper(I_Y) += (filext > 1) ? filext : 1;
				cproseq->upper(I_Y) += 1;
			}
		} else {
ifelse(NDIM,2,@
			*cfilseq = (*cfilseq)*f.domain_R3;
@,NDIM,3,@
			*cfilseq = (*cfilseq)*grow(f.domain_R3,periodic);
@)
			if(cintseq->lower(I_Y) == 1) cfilseq->lower(I_Y) -= 1;
			*cproseq = *cfilseq;
		}
		*(cextseq++) += *cfilseq;
		cintseq++;
		cfilseq++;
		cproseq++;
	}
//
	delete [] al.interiors;
	delete [] al.extents;
	delete [] al.fills;
	delete [] al.coarse_interiors;
	delete [] al.coarse_extents;
	delete [] al.coarse_fills;
	delete [] al.coarse_prolong;
	al.mapping = map_new;
	al.interiors = int_new;
	al.extents = ext_new;
	al.fills = fil_new;
	al.coarse_interiors = cint_new;
	al.coarse_extents = cext_new;
	al.coarse_fills = cfil_new;
	al.coarse_prolong = cpro_new;
	al.Ninteriors = regions;
	al.baseP_R1 = f.baseP_R1;
	al.baseP_R2 = f.baseP_R2;
	al.baseP_R3 = f.baseP_R3;
//
//------------------------------------------------------------------------------
//	Locate all duplicate fill region points.
ifelse(NDIM,3,@
	const int N = al.domain_R3.extents(I_W);
	Point3 up(0); up(I_W) = N;
	Point3 dn(0); dn(I_W) = -N;
	const Region3 PhiMask = al.domain_R1 + al.domain_R2 + al.domain_R3;
@)
	FillList *fl = FillListNULL;
	al.dup_cnt = 0;
	for(int l = 0; l < regions; l++) {
		for(int m = l+1; m < regions; m++) {
			RegionX ol = al.fills[l]*al.fills[m];
ifelse(NDIM,2,@
			for_2(i,j,ol)
				if(i<al.interiors[l].lower(0) || j<al.interiors[l].lower(1) ||
					i>al.interiors[l].upper(0) || j>al.interiors[l].upper(1)) {
					if(i<al.interiors[m].lower(0) || j<al.interiors[m].lower(1) ||
						i>al.interiors[m].upper(0) || j>al.interiors[m].upper(1)) {
						Point2 addpt(i,j);
						FillList ** test = &fl;
						do {
							if(*test == FillListNULL) {
								(*test) = new FillList(addpt);
								++al.dup_cnt;
								break;
							}
							if((*test)->where == addpt) break;
							test = &((*test)->next);
						}
						while(1);
					}
				}
			end_for
@,NDIM,3,@
// do not allow phi-periodic fill points to be counted
			ol *= PhiMask;
//
			for_3(i,j,k,ol)
				if(i<al.interiors[l].lower(0) || j<al.interiors[l].lower(1) ||
					k<al.interiors[l].lower(2) || i>al.interiors[l].upper(0) ||
					j>al.interiors[l].upper(1) || k>al.interiors[l].upper(2)) {
					if(i<al.interiors[m].lower(0) || j<al.interiors[m].lower(1) ||
						k<al.interiors[m].lower(2) || i>al.interiors[m].upper(0) ||
						j>al.interiors[m].upper(1) || k>al.interiors[m].upper(2)) {
						Point3 addpt(i,j,k);
						FillList ** test = &fl;
						do {
							if(*test == FillListNULL) {
								(*test) = new FillList(addpt);
								++al.dup_cnt;
								break;
							}
							if((*test)->where == addpt) break;
							test = &((*test)->next);
						}
						while(1);
					}
				}
			end_for
@)
		}
ifelse(NDIM,3,@
// Handle phi-periodic fill points
		for(m = l; m < regions; m++) {
			Region3 ol = al.fills[l]*shift(al.fills[m],up)*PhiMask;
			Region3 pints = shift(al.interiors[m],up);
			for_3(i,j,k,ol)
				if(i<al.interiors[l].lower(0) || j<al.interiors[l].lower(1) ||
					k<al.interiors[l].lower(2) || i>al.interiors[l].upper(0) ||
					j>al.interiors[l].upper(1) || k>al.interiors[l].upper(2)) {
					if(i<pints.lower(0) || j<pints.lower(1) ||
						k<pints.lower(2) || i>pints.upper(0) ||
						j>pints.upper(1) || k>pints.upper(2)) {
						Point3 addpt(i,j,k);
						FillList ** test = &fl;
						do {
							if(*test == FillListNULL) {
								(*test) = new FillList(addpt);
								++al.dup_cnt;
								break;
							}
							if((*test)->where == addpt) break;
							test = &((*test)->next);
						}
						while(1);
					}
				}
			end_for
			ol = al.fills[l]*shift(al.fills[m],dn)*PhiMask;
			pints = shift(al.interiors[m],dn);
			for_3(i,j,k,ol)
				if(i<al.interiors[l].lower(0) || j<al.interiors[l].lower(1) ||
					k<al.interiors[l].lower(2) || i>al.interiors[l].upper(0) ||
					j>al.interiors[l].upper(1) || k>al.interiors[l].upper(2)) {
					if(i<pints.lower(0) || j<pints.lower(1) ||
						k<pints.lower(2) || i>pints.upper(0) ||
						j>pints.upper(1) || k>pints.upper(2)) {
						Point3 addpt(i,j,k);
						FillList ** test = &fl;
						do {
							if(*test == FillListNULL) {
								(*test) = new FillList(addpt);
								++al.dup_cnt;
								break;
							}
							if((*test)->where == addpt) break;
							test = &((*test)->next);
						}
						while(1);
					}
				}
			end_for
		}
// End of phi-periodic fill point handler
@)
	}
	FillList * fin = FillListNULL;
	while(fl != FillListNULL) {
		FillList * check = fl;
		fl = fl->next;
ifelse(NDIM,2,@
		const int i = check->where(0);
		const int j = check->where(1);
		for(int l = 0; l < regions; l++) {
			if(i >= al.fills[l].lower(0) && i <= al.fills[l].upper(0) &&
				j >= al.fills[l].lower(1) && j <= al.fills[l].upper(1)) {
				if(i >= al.interiors[l].lower(0) && i <= al.interiors[l].upper(0) &&
					j >= al.interiors[l].lower(1) && j <= al.interiors[l].upper(1)) {
					check->count = 0;
					--al.dup_cnt;
					break;
				} else {
					++(check->count);
				}
			}
		}
@,NDIM,3,@
		const int i = check->where(0);
		const int j = check->where(1);
		const int k = check->where(2);
		for(int l = 0; l < regions; l++) {
			if(i >= al.fills[l].lower(0) && i <= al.fills[l].upper(0) &&
				j >= al.fills[l].lower(1) && j <= al.fills[l].upper(1) &&
				k >= al.fills[l].lower(2) && k <= al.fills[l].upper(2)) {
				if(i >= al.interiors[l].lower(0) && i <= al.interiors[l].upper(0) &&
					j >= al.interiors[l].lower(1) && j <= al.interiors[l].upper(1) &&
					k >= al.interiors[l].lower(2) && k <= al.interiors[l].upper(2)) {
					check->count = 0;
					--al.dup_cnt;
					break;
				} else {
					++(check->count);
				}
			}
//	handle phi-periodic interior points
			Region3 pfu = shift(PhiMask*al.fills[l],up);
			Region3 piu = shift(al.interiors[l],up);
			if(i >= pfu.lower(0) && i <= pfu.upper(0) &&
				j >= pfu.lower(1) && j <= pfu.upper(1) &&
				k >= pfu.lower(2) && k <= pfu.upper(2)) {
				if(i >= piu.lower(0) && i <= piu.upper(0) &&
					j >= piu.lower(1) && j <= piu.upper(1) &&
					k >= piu.lower(2) && k <= piu.upper(2)) {
					check->count = 0;
					--al.dup_cnt;
					break;
				} else {
					++(check->count);
				}
			}
			Region3 pfd = shift(PhiMask*al.fills[l],dn);
			Region3 pid = shift(al.interiors[l],dn);
			if(i >= pfd.lower(0) && i <= pfd.upper(0) &&
				j >= pfd.lower(1) && j <= pfd.upper(1) &&
				k >= pfd.lower(2) && k <= pfd.upper(2)) {
				if(i >= pid.lower(0) && i <= pid.upper(0) &&
					j >= pid.lower(1) && j <= pid.upper(1) &&
					k >= pid.lower(2) && k <= pid.upper(2)) {
					check->count = 0;
					--al.dup_cnt;
					break;
				} else {
					++(check->count);
				}
			}
//	End of phi-periodic interior points handler
		}
@)
		if(check->count != 0) {
			check->next = fin;
			fin = check;
		} else {
			delete check;
		}
	}
	al.dup_pts = new PointX[al.dup_cnt];
	al.dup_value = new double[NDUPFIELDS*al.dup_cnt];
	al.dup_norm = new double[al.dup_cnt];
	for(k = 0; k < al.dup_cnt; k++) {
		al.dup_pts[k] = fin->where;
		al.dup_norm[k] = 1.0/fin->count;
		FillList *remove = fin;
		fin = fin->next;
		delete remove;
	}
//
//------------------------------------------------------------------------------
	initialize_level(New_Level,1.0,1,0,Old_Level);
//
//------------------------------------------------------------------------------
//	Delete the old discretization if it exists
	delete Old_Level;
//------------------------------------------------------------------------------
	return(1);
}


void cadez_hierarchy::refine_solution(const int lev) {
	cadez_discretization & f = *(level[lev]);
	cadez_discretization & c = *(level[lev-1]);
	for_all_1(i,f.PsiC)
		for_1(j,c.PsiD)
			f.PsiC(i).copy(c.PsiD(j),c.fills[j]*f.coarse_fills[i]);
		end_for
	end_for_all
	for_all_1(l,f.PsiD)
ifelse(NDIM,2,@
		ctof3x3(f.PsiRes(l).data(),f.PsiC(l).data(),
					(int *)(f.extents+l),(int *)(f.coarse_prolong+l),
					(int *)(f.coarse_extents+l),(int *)(f.coarse_fills+l));
		const Grid2(double) & src = f.PsiRes(l);
		Grid2(double) & dst = f.PsiD(l);
		BeginFastIndex2(src, const double);
		BeginFastIndex2(dst, double);
		for_2(i,j,f.fills[l])
			FastIndex2(dst,i,j) = FastIndex2(src,i,j);
		end_for
		EndFastIndex2(src);
		EndFastIndex2(dst);
@,NDIM,3,@
		ctof3x3x3(f.PsiRes(l).data(),f.PsiC(l).data(),
					(int *)(f.extents+l),(int *)(f.coarse_prolong+l),
					(int *)(f.coarse_extents+l),(int *)(f.coarse_fills+l));
		const Grid3(double) & src = f.PsiRes(l);
		Grid3(double) & dst = f.PsiD(l);
		BeginFastIndex3(src, const double);
		BeginFastIndex3(dst, double);
		for_3(i,j,k,f.fills[l])
			FastIndex3(dst,i,j,k) = FastIndex3(src,i,j,k);
		end_for
		EndFastIndex3(src);
		EndFastIndex3(dst);
@)
		f.PsiRes(l).fill(0.0);
	end_for_all_no_sync
	f.fill_patch(f.PsiD);
//
//	Set duplicate fill points
	if(f.dup_cnt > 0) {
ifelse(NDIM,3,@
	const int N = f.domain_R3.extents(I_W);
@)
		for(int m = 0; m < NDUPFIELDS*f.dup_cnt; m++) f.dup_value[m] = 0.0;
		for_all_1(l,f.PsiD)
//cout << "\nSetting From Patch " << l << endl;
			const int l0 = f.fills[l].lower(0);
			const int l1 = f.fills[l].lower(1);
			ifelse(NDIM,3,const int l2 = f.fills[l].lower(2);)
			const int h0 = f.fills[l].upper(0);
			const int h1 = f.fills[l].upper(1);
			ifelse(NDIM,3,const int h2 = f.fills[l].upper(2);)
			for(int m = 0; m < f.dup_cnt; m++) {
//cout << "\nPoint " << f.dup_pts[m] << " | ";
				const int i = f.dup_pts[m](0);
				const int j = f.dup_pts[m](1);
				ifelse(NDIM,3,const int k = f.dup_pts[m](2);)
ifelse(NDIM,2,@
				if(i >= l0 && i <= h0 && j >= l1 && j <= h1)
					f.dup_value[NDUPFIELDS*m] += f.PsiD(l)(i,j);
@,NDIM,3,@
				if(i >= l0 && i <= h0 && j >= l1 && j <= h1 && k >= l2 && k <= h2) {
					f.dup_value[NDUPFIELDS*m] += f.PsiD(l)(i,j,k);
//cout << f.PsiD(l)(i,j,k) << " : ";
				}
@)
			}
//cout << endl;
		end_for_all_no_sync
		mpReduceAdd(f.dup_value,NDUPFIELDS*f.dup_cnt);
		for(m = 0; m < f.dup_cnt; m++) {
			f.dup_value[NDUPFIELDS*m] *= f.dup_norm[m];
		}
		for_all_1(l,f.PsiD)
//cout << "\nFilling Patch " << l << endl;
			const int l0 = f.fills[l].lower(0);
			const int l1 = f.fills[l].lower(1);
			ifelse(NDIM,3,const int l2 = f.fills[l].lower(2);)
			const int h0 = f.fills[l].upper(0);
			const int h1 = f.fills[l].upper(1);
			ifelse(NDIM,3,const int h2 = f.fills[l].upper(2);)
			for(int m = 0; m < f.dup_cnt; m++) {
				const int i = f.dup_pts[m](0);
				const int j = f.dup_pts[m](1);
				ifelse(NDIM,3,const int k = f.dup_pts[m](2);)
ifelse(NDIM,2,@
				if(i >= l0 && i <= h0 && j >= l1 && j <= h1)
					f.PsiD(l)(i,j) = f.dup_value[NDUPFIELDS*m];
@,NDIM,3,@
//				if(i >= l0 && i <= h0 && j >= l1 && j <= h1 && k >= l2 && k <= h2)
//					f.PsiD(l)(i,j,k) = f.dup_value[NDUPFIELDS*m];
//	Make sure all phi-periodic duplicate points are filled
				if(j >= l1 && j <= h1 && k >= l2 && k <= h2) {
					if(i >= l0 && i <= h0) {
						f.PsiD(l)(i,j,k) = f.dup_value[NDUPFIELDS*m];
//cout << "Setting (" << i << "," << j << "," << k << ") with " <<
//			 f.dup_value[NDUPFIELDS*m] << endl;
					}
					if(i+N >= l0 && i+N <= h0) {
						f.PsiD(l)(i+N,j,k) = f.dup_value[NDUPFIELDS*m];
//cout << "Setting (" << i+N << "," << j << "," << k << ") with " <<
//			 f.dup_value[NDUPFIELDS*m] << endl;
					}
					if(i-N >= l0 && i-N <= h0) {
						f.PsiD(l)(i-N,j,k) = f.dup_value[NDUPFIELDS*m];
//cout << "Setting (" << i-N << "," << j << "," << k << ") with " <<
//			 f.dup_value[NDUPFIELDS*m] << endl;
					}
				}
//	Done filling all phi-periodic duplicate points
@)
			}
		end_for_all
	}
	f.sync_fields();
}


void Balance_Partitions(RegionList *& regions, RegionX *&interiors,
								int *&assignments, int & nblocks,
								const int processors, const PointX & minsize) {
//
// Calculate the average block size to be assigned to a processor
	int total = 0;
	for(RegionList *p = regions; p != RegionListNULL; p = p->next)
		total += p->region.size();
	const int average = total/processors;
	int n = nblocks;
	SliceAndDice(regions, n, average, minsize);
	if(n > 1) SortRegions(regions,n);
	interiors = new RegionX[n];
	assignments = new int[n];
	int *work = new int[processors];
	for(int i=0; i < processors; i++) work[i] = 0;
	int rcount = 0;
	for(RegionList *pb = regions; pb != RegionListNULL; pb = pb->next) {
		const int size = pb->region.size();
		int diff = average-(work[0]+size), proc = 0;
		for(int pp = 1; pp < processors; pp++) {
			const int d = average-(work[pp]+size);
			if(((diff > 0) && (d >= 0) && (d < diff)) ||
					((diff < 0) && (d > diff))){
				diff = d;
				proc = pp;
			}
		}
		work[proc] += size;
		assignments[rcount] = proc;
		interiors[rcount++] = pb->region;
	}
	delete [] work;
	nblocks = n;
	for(int j=0; j < n; j++) {
		RegionList *byebye = regions;
		regions = regions->next;
		delete byebye;
	}
}


static int find_interior_zero(int& xcut, const int x0, const int x1,
						const Array1(int)& sig, const int minsize) {
	const int a = x0+minsize-1;
	const int b = x1-minsize;
	const int m = (x0+x1)/2;
	for(int d = 0; (m-d >= a) && (m+d <= b); d++) {
		if(sig(m-d) == 0) { xcut = m-d; return(1); }
		if(sig(m+d) == 0) { xcut = m+d; return(1); }
	}
	return(0);
}

static void cut_laplacian(int& cut, const int low, const int high,
									const Array1(int)& sig, const int minsize) {
	const int a = MAX(low+minsize-1, low+1);
	const int b = MIN(high-minsize, high-2);
	int lastlap = sig(a-1)-2*sig(a)+sig(a+1);
	int maxzero = 0;
	int loczero = (low+high)/2;
	for(int l = a+1; l <= b+1; l++) {
		const int newlap = sig(l-1)-2*sig(l)+sig(l+1);
		if(((newlap < 0) && (lastlap >= 0)) ||
				((newlap >= 0) && (lastlap < 0))) {
			int delta = newlap - lastlap;
			if(delta < 0) delta = -delta;
			if(delta > maxzero) {
				loczero = l-1;
				maxzero = delta;
			}
		}
		lastlap = newlap;
	}
	cut = loczero;
}

static int split_region(RegionX& left, RegionX& right, const RegionX& region,
					const Array1(int) sigs[NDIM], const PointX& minsize) {
//
//	Find the width of the region, sort to get the longest, and see what can
//	be split
	const PointX lower = region.lower();
	const PointX upper = region.upper();
	const PointX width = upper-lower+1;
//
	ifelse(NDIM,2,Point2 sorted(0,1);)
//	ifelse(NDIM,3,Point3 sorted(0,1,2);)
	ifelse(NDIM,3,Point3 sorted(2,1,0);)
	for(int ii = 0; ii < NDIM-1; ii++)
		for(int jj = ii+1; jj < NDIM; jj++)
			if(width(sorted(ii)) < width(sorted((jj)))) {
				const int tmp = sorted(ii);
				sorted(ii) = sorted(jj);
				sorted(jj) = tmp;
			}
//
//	See how many dimensions are splitable according to minsize
//	Do not cut a side if the aspect ratio is already out of wack
	for(int nsplit = 0; nsplit < NDIM; nsplit++)
		if((width(sorted(nsplit)) < 2*minsize(sorted(nsplit))) ||
			(width(sorted(nsplit)) < width(sorted(0))/2)) break;
	if(nsplit == 0) return(FALSE);
//
//	Attempt to cut using a zero interior point from widest to shortest
	int cut, s, d;
	for (d = 0; d < nsplit; d++) {
		s = sorted(d);
		if(find_interior_zero(cut,lower(s),upper(s),sigs[s],minsize(s))) break;
	}
//
//	If no split was made, then split using the Laplacian on the longest side
	if(d == nsplit) {
		s = sorted(0);
		cut_laplacian(cut, lower(s), upper(s), sigs[s], minsize(s));
	}
//
//	Using the cut information, generate the two subregions
	PointX left_upper  = upper;
	PointX right_lower = lower;
	left_upper(s)  = cut;
	right_lower(s) = cut+1;
	left  = RegionX(lower,       left_upper);
	right = RegionX(right_lower, upper     );
//
	return(1);
}

static RegionList *RegridLevel(int & regions, cadez_discretization & lev, 
			const RegionX & bbox, const int threshold, const PointX & minsize) {
//
//	Compute "signature" for the flagged region
	Array1(int) sigs[NDIM];
	for(int d = 0; d < NDIM; d++) {
		sigs[d].resize(Region1(bbox.lower(d),bbox.upper(d)));
		sigs[d].fill(0);
		for_all_1(i,lev.Refine_Mask)
			RegionX local = lev.interiors[i]*bbox;
			if(local.empty()) continue;
			const GridX(double) & grid = lev.Refine_Mask(i);
ifelse(NDIM,2,@
			switch (d) {
				case 0:
					BeginFastIndex2(grid, const double);
					for_2(j,k,local)
						if(FastIndex2(grid,j,k) == 0.0) sigs[d](j) += 1;
					end_for
					EndFastIndex2(grid);
					break;
				case 1:
					BeginFastIndex2(grid, const double);
					for_2(j,k,local)
						if(FastIndex2(grid,j,k) == 0.0) sigs[d](k) += 1;
					end_for
					EndFastIndex2(grid);
					break;
			}
@,NDIM,3,@
			switch (d) {
				case 0:
					BeginFastIndex3(grid, const double);
					for_3(j,k,l,local)
						if(FastIndex3(grid,j,k,l) == 0.0) sigs[d](j) += 1;
					end_for
					EndFastIndex3(grid);
					break;
				case 1:
					BeginFastIndex3(grid, const double);
					for_3(j,k,l,local)
						if(FastIndex3(grid,j,k,l) == 0.0) sigs[d](k) += 1;
					end_for
					EndFastIndex3(grid);
					break;
				case 2:
					BeginFastIndex3(grid, const double);
					for_3(j,k,l,local)
						if(FastIndex3(grid,j,k,l) == 0.0) sigs[d](l) += 1;
					end_for
					EndFastIndex3(grid);
					break;
			}
@)
		end_for_all_no_sync
		mpReduceAdd(sigs[d].data(),sigs[d].size());
	}
	int nflagged = 0;
	for_1(i,sigs[0])
		nflagged += sigs[0](i);
	end_for
//
//	Find the minimum box for the region and also the efficiency.  If the
//	efficiency is poor then split this region into subregions.
	RegionX inner_bounds;
	for(d = 0; d < NDIM; d++) {
		int & x0 = inner_bounds.lower(d);
		int & x1 = inner_bounds.upper(d);
		Array1(int) & sig = sigs[d];
		const int ms = minsize(d);
		const int a = x0 = sig.lower(0);
		const int b = x1 = sig.upper(0);
		int width = b - a + 1;
		if(width > ms) {
			while((x0 <= x1) && (sig(x0) == 0)) (x0)++;
			while((x1 >= x0) && (sig(x1) == 0)) (x1)--;
			if((width = (x1 - x0 + 1)) < ms) {
				const int pad = ms - width;
				x0 -= (pad+1)/2;
				if(x0 < a) x0 = a;
				x1 = x0 + ms - 1;
				if(x1 > b) x1 = b;
			}
		}
	}
	const int npoints = inner_bounds.size();
	const int eff = (npoints == 0 ? 100 : (100*nflagged)/npoints);
//
	if(inner_bounds.empty()) return(RegionListNULL);
//
// Try to split the region.
#define COMBINE_THRESHOLD 90
	RegionList *covering;
	RegionX a, b;
	if((eff < threshold) && (split_region(a,b,inner_bounds,sigs,minsize))) {
		int n_a, n_b;
		RegionList *cover_a = RegridLevel(n_a, lev, a, threshold, minsize);
		RegionList *cover_b = RegridLevel(n_b, lev, b, threshold, minsize);
		int combined = cover_a->region.size() + cover_b->region.size();
		if((n_a == 1) && (n_b == 1) &&
				(100*combined/npoints >= COMBINE_THRESHOLD)) {
		//	If we were not able to split the region, then put it back together.
		//	If the efficiency of the combined block is not that different from
		//	the broken block, just return the original block.
			cover_a->region = inner_bounds;
			covering = cover_a;
			regions = 1;
			delete cover_b;
		} else {
		//	We did do some refinements, so append the lists and return.
			regions = n_a + n_b;
			for(RegionList *p = cover_a; p->next; p = p->next);
			covering = cover_a;
			p->next = cover_b;
		}
	} else {
	//	We have already meet threshold requirements or splitting would violate
	//	the minsize requirement for the region.
		covering = new RegionList;
		covering->next = RegionListNULL;
		covering->region = inner_bounds;
		regions = 1;
	}
	return(covering);
}

static void Burstem(const RegionX& bursty, const RegionX& solid,
							RegionList*& list, const int dimension) {
//
//	Set up the lower and upper bounds of the regions for ease of access
	PointX burstl = bursty.lower();
	PointX bursth = bursty.upper();
	const PointX& solidl = solid.lower();
	const PointX& solidh = solid.upper();
//
//	Break up the bursty region against the solid region along
//	low dimensions first
	for(int d = 0; d < dimension; d++) {
		if(bursth(d) > solidh(d)) {
			PointX newl = burstl;
			newl(d) = solidh(d) + 1;
			RegionList *p = new RegionList;
			p->region.setregion(newl, bursth);
			p->next = list;
			list = p;
			bursth(d) = solidh(d);
		}
		if(burstl(d) < solidl(d)) {
			PointX newh = bursth;
			newh(d) = solidl(d) - 1;
			RegionList *p = new RegionList;
			p->region.setregion(burstl, newh);
			p->next = list;
			list = p;
			burstl(d) = solidl(d);
		}
	}
}

static RegionList *CoalesceNestedList(RegionList *nested) {
//
//	Start coalescing on the highest dimension of the lists and work down
	RegionList *canon = RegionListNULL;
	RegionList *other = nested;
	for(int d = NDIM-1; d >= 0; d--) {
	//
	//	While there are non-canonical regions, pick somebody off of the other list
		while(other != RegionListNULL) {
			RegionList *tryme = other;
			other = other->next;
		//
		//	Pick somebody off of the canon list and compare against tryme
			RegionList **prev = &canon;
			while(*prev != RegionListNULL) {
				RegionList *andme = *prev;
			//
			//	See if these two guys are in canonical order - if not break out
				const PointX& al = andme->region.lower();
				const PointX& ah = andme->region.upper();
				const PointX& bl = tryme->region.lower();
				const PointX& bh = tryme->region.upper();
				int coalesce = TRUE;
				for(int du = d+1; du < NDIM; du++)
					if ((al(du) != bl(du)) || (ah(du) != bh(du))) {
						coalesce = FALSE;
						break;
					}
				if(coalesce) {
					if((bl(d) > ah(d)+1) || (bh(d) < al(d)-1))
						coalesce = FALSE;
					else
						for(int dl = 0; dl < d; dl++)
							if((bl(dl) > ah(dl)) || (bh(dl) < al(dl))) {
								coalesce = FALSE;
								break;
							}
				}
				if(coalesce) break;
				prev = &((*prev)->next);
			}
		//
		//	If we are at the end of the canonical list, then just add
			if(*prev == RegionListNULL) {
				tryme->next = canon;
				canon = tryme;
			} else {
		//
		//	Otherwise, we need to burst tryme and andme and put on the list
		//	Unlink the offending box from the canonical list
				RegionList *andme = *prev;
				*prev = (*prev)->next;
		//	Generate the intersection lower and upper bounds
				const PointX& bl = tryme->region.lower();
				const PointX& bh = tryme->region.upper();
				PointX il = andme->region.lower();
				PointX ih = andme->region.upper();
				for(int dl = 0; dl < d; dl++) {
					if (il(dl) < bl(dl)) il(dl) = bl(dl);
					if (ih(dl) > bh(dl)) ih(dl) = bh(dl);
				}
				if(bl(d) < il(d)) il(d) = bl(d);
				if(bh(d) > ih(d)) ih(d) = bh(d);
			// Add this new puppy to the list of other regions
				RegionList *incommon = new RegionList;
				incommon->region.setregion(il, ih);
				incommon->next = other;
				other = incommon;
			//	We have the intersection --- burst everything else
				if(d > 0) {
					Burstem(tryme->region, incommon->region, other, d);
					Burstem(andme->region, incommon->region, other, d);
				}
			//	Free the two regions that we just combined
				delete tryme;
				delete andme;
			}
		}
	//
	//	All of the regions are now in proper canonical order for dimensions
	//	greater and equal to d --- start over again from the beginning
		other = canon;
		canon = RegionListNULL;
	}
	return(other);
}

static RegionList *CheckNesting(RegionX& region, cadez_discretization & lev) {
//
//	Intersect region with all subregions in flags and store intersections
	RegionList *nested = RegionListNULL;
	for_1(i, lev.Refine_Mask)
		RegionX nest = region * lev.interiors[i];
		if(!nest.empty()) {
			RegionList *newnest = new RegionList;
			newnest->region = nest;
			newnest->next = nested;
			nested = newnest;
		}
	end_for
//	Coalesce this box list back together if possible and return
	return(CoalesceNestedList(nested));
}

static void EnsureNesting(RegionList*& list, int& regions,
									cadez_discretization & lev) {
	RegionList *nested = RegionListNULL;
	regions = 0;
	while(list != RegionListNULL) {
		RegionList *nest = CheckNesting(list->region, lev);
		regions++;
		for (RegionList *p = nest; p->next; p = p->next, regions++);
		p->next = nested;
		nested = nest;
		RegionList *byebye = list;
		list = list->next;
		delete byebye;
	}
	list = nested;
}

static int count_flagged(const RegionX& region, cadez_discretization & lev) {
	Array1(int) sig(Region1(region.lower(0), region.upper(0)));
	sig.fill(0);
	for_all_1(i,lev.Refine_Mask)
		RegionX local = lev.interiors[i]*region;
		if(local.empty()) continue;
		const GridX(double) & grid = lev.Refine_Mask(i);
		BeginFastIndexNDIM(grid, const double);
ifelse(NDIM,2,@
		for_2(j,k,local)
			if(FastIndex2(grid,j,k) == 0.0) sig(j) += 1;
		end_for
@,NDIM,3,@
		for_3(j,k,l,local)
			if(FastIndex3(grid,j,k,l) == 0.0) sig(j) += 1;
		end_for
@)
		EndFastIndexNDIM(grid);
	end_for_all_no_sync
	mpReduceAdd(sig.data(),sig.size());
	int nflagged = 0;
	for_1(i, sig)
		nflagged += sig(i);
	end_for
	return(nflagged);
}

static int SliceOK(const RegionX& region, RegionX& fragment, RegionX& rest,
							const int average, const PointX& minsize) {
//
//	Find the width of the region, sort to get the longest,
//	and see what can be split
	const PointX lower = region.lower();
	const PointX upper = region.upper();
	const PointX width = upper-lower+1;
	ifelse(NDIM,2,Point2 sorted(0,1);)
//	ifelse(NDIM,3,Point3 sorted(0,1,2);)
	ifelse(NDIM,3,Point3 sorted(2,1,0);)
	for(int ii = 0; ii < NDIM-1; ii++)
		for(int jj = ii+1; jj < NDIM; jj++)
		if(width(sorted(ii)) < width(sorted((jj)))) {
			const int tmp = sorted(ii);
			sorted(ii) = sorted(jj);
			sorted(jj) = tmp;
		}
//
//	Try each dimension in turn --- check if we can slit this dimension
	for(int d = 0; d < NDIM; d++) {
		const int s = sorted(d);
	//
	//	Calculate where we would have to cut to render average size region
		int size = 1;
		for(int i = 0; i < NDIM; i++)
			if(i != s) size *= width(i);
		int cut = ROUND(average, size);
		if(cut < minsize(s)) cut = minsize(s);
		if((width(s)-cut) >= minsize(s)) {
			PointX left_upper  = upper;
			PointX right_lower = lower;
			const int cuthere = lower(s)+cut-1;
			left_upper(s)  = cuthere;
			right_lower(s) = cuthere+1;
			fragment.setregion(lower, left_upper);
			rest.setregion(right_lower, upper);
			return(1);
		}
	}
	return(0);
}

static void SliceAndDice(RegionList*& regions, int& n, const int average,
									const PointX& minsize) {
   RegionList *fragmented = RegionListNULL;
//	Fragment the blocks in list regions and place them on the fragmented list
	while(regions != RegionListNULL) {
		RegionList *frag = regions;
		regions = regions->next;
	//	If the size of regions is greater than average, try to split it
		while(frag->region.size() > average) {
			RegionX fragment, rest;
			if(!SliceOK(frag->region, fragment, rest, average, minsize)) break;
			RegionList *newfrag = new RegionList;
			newfrag->region = fragment;
			newfrag->next = fragmented;
			fragmented = newfrag;
			n++;
			frag->region = rest;
		}
	//	Place the remaining fragment on the fragmented list
		frag->next = fragmented;
		fragmented = frag;
	}
	regions = fragmented;
}

static void heapify(RegionList **heap, const int i, const int j) {
	const int l = 2*i+1;
	const int r = l+1;
	int s = i;
	if((l < j) && (heap[s]->region.size() > heap[l]->region.size())) s = l;
	if((r < j) && (heap[s]->region.size() > heap[r]->region.size())) s = r;
	if(s != i) {
		RegionList *tmp = heap[s];
		heap[s] = heap[i];
		heap[i] = tmp;
		heapify(heap, s, j);
	}
}

static void SortRegions(RegionList*& regions, const int nblocks) {
//
//	Store pointers to the blocks in an array which will be heapsorted
	RegionList **heap = new RegionList *[nblocks];
	RegionList *p = regions;
	for(int i = 0; i < nblocks; i++, p = p->next)
		heap[i] = p;
//
//	Do the heapsort thing (hey, I'm using Aho, Hopcroft, and Ullman --- Cool!)
	for(int j = nblocks/2-1; j >= 0; j--)
		heapify(heap, j, nblocks);
	for(int k = nblocks-1; k >= 1; k--) {
		RegionList *tmp = heap[0];
		heap[0] = heap[k];
		heap[k] = tmp;
		heapify(heap, 0, k);
	}
//
//	Reform the linked list and delete the heap area
	for(int l = 0; l < nblocks-1; l++)
		heap[l]->next = heap[l+1];
	heap[l]->next = RegionListNULL;
	regions = heap[0];
	delete [] heap;
}
